#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : convolve

"""
This function convolves two numpy arrays (1D, 2D, 3D).
"""



# %% Libraries
from arrlp import correlate
import numpy as np



# %% Function
def convolve(arr, k=None, /, out=None, **kwargs) :
    '''
    This function convolves two numpy arrays (1D, 2D, 3D).
    
    Parameters
    ----------
    arr : numpy.ndarray
        first array to correlate.
    k : numpy.ndarray
        second array to correlate, if is None is auto created.
    out : numpy.ndarray
        array where to put correlation, same shape as arr, if None is auto initialized.
    kwargs : dict
        Parameters to pass to arrlp.correlate function.

    Returns
    -------
    out : numpy.ndarray
        Correlated array.

    Raises
    ------
    TypeError
        if output shape does not match array shape.

    Examples
    --------
    >>> from arrlp import correlate
    ...
    >>> convolve(img, sigma=5) # Gaussian convolve
    >>> convolve(img, window=10) # Mask convolve
    >>> convolve(img, wl=640, NA=1.5) # Airy convolve
    '''

    if k is not None :
        for dim in range(np.ndim(k)) :
            k = np.flip(k, axis=dim)
    return correlate(arr, k, out, **kwargs)
    



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : correlate

"""
This function correlates two numpy arrays (1D, 2D, 3D).
"""



# %% Libraries
from arrlp import kernel
from numba import njit, prange
import numpy as np
import astropy



# %% Function
def correlate(arr, k=None, /, out=None, *, method="mean", **kernel_kwargs) :
    '''
    This function correlates two numpy arrays (1D, 2D, 3D).
    
    Parameters
    ----------
    arr : numpy.ndarray
        first array to correlate.
    k : numpy.ndarray
        second array to correlate, if is None is auto created.
    out : numpy.ndarray
        array where to put correlation, same shape as arr, if None is auto initialized.
    method : str
        define function to use for correlation. Should be in ["astro", "mean", "std", "min", "max", "sum", "prod"].
    kernel_kwargs : dict
        Parameters passed to kernel function for auto-creation.

    Returns
    -------
    out : numpy.ndarray
        Correlated array.

    Raises
    ------
    TypeError
        if output shape does not match array shape.

    Examples
    --------
    >>> from arrlp import correlate
    ...
    >>> correlate(img, sigma=5) # Gaussian correlation
    >>> correlate(img, window=10) # Mask correlation
    >>> correlate(img, wl=640, NA=1.5) # Airy correlation
    '''

    # Method
    ndims = len(np.shape(arr))
    method = f"{method}{ndims}"

    # Kernel
    if k is None :
        k = kernel(ndims, **kernel_kwargs)
    
    # Output
    if out is None :
        out = np.empty_like(arr)
    if np.shape(out) != np.shape(arr) :
        raise TypeError('Output shape does not match array shape')
    
    # Function
    function = function_factory(method)
    function(arr, k, out)
    
    return out



_cached = {}
funcs = {
    'mean' : np.nanmean,
    'std' : np.nanstd,
    'min' : np.nanmin,
    'max' : np.nanmax,
    'sum' : np.nansum,
    'prod' : np.nanprod,
    }

def function_factory(method) :
    function = _cached.get(method, None)
    if function is not None :
        return function

    if method.startswith("astro") :
        if method.endswith("1") :
            def invert(k) :
                return k[::-1]
        elif method.endswith("2") :
            def invert(k) :
                return k[::-1, ::-1]
        elif method.endswith("3") :
            def invert(k) :
                return k[::-1, ::-1, ::-1]
        else :
            raise SyntaxError(f'Correlation error cannot be {method}')

        def function(arr, k, out, *, kernel_limit=2**12) :
            k = invert(k)
            if k.size > kernel_limit :
                out[:] = astropy.convolution.convolve_fft(arr, k, normalize_kernel=False)
            else :
                out[:] = astropy.convolution.convolve(arr, k, boundary='extend', normalize_kernel=False)
        _cached[method] = function
        return function

    func = funcs[method[:-1]]

    if method.endswith("1") :
        @njit(parallel=True)
        def function(arr, k, out):
            xlim, = np.shape(arr)
            xp, = np.shape(k)
            xp = int(xp / 2)
            sum = np.sum(k)
            sizep = k.size
            for i in prange(xlim):
                x = i
                xmin, xmax = max(0, x - xp), min(xlim, x + xp + 1)
                n = (xmax - xmin)
                if n != sizep :
                    ksum = np.sum(k[xmin - x + xp: xmax - x + xp])
                    if ksum != 0 :
                        n = n * sum / ksum
                crop = arr[xmin:xmax] * k[xmin - x + xp: xmax - x + xp] * n
                out[x] = func(crop)

    elif method.endswith("2") :
        
        @njit(parallel=True)
        def function(arr, k, out):
            ylim, xlim = np.shape(arr)
            yp, xp = np.shape(k)
            yp, xp = int(yp / 2),int(xp / 2)
            sum = np.sum(k)
            sizep = k.size
            for y in prange(ylim):
                for x in range(xlim) :
                    ymin, ymax = max(0, y - yp), min(ylim,y + yp + 1)
                    xmin, xmax = max(0, x - xp), min(xlim,x + xp + 1)
                    n = (ymax - ymin) * (xmax - xmin)
                    if n != sizep :
                        ksum = np.sum(k[ymin - y + yp: ymax - y + yp, xmin - x + xp: xmax - x + xp])
                        if ksum != 0 :
                            n = n * sum / ksum
                    crop = arr[ymin:ymax, xmin:xmax] * k[ymin - y + yp: ymax - y + yp, xmin - x + xp: xmax - x + xp] * n
                    out[y, x] = func(crop)

    elif method.endswith("3") :
        
        @njit(parallel=True)
        def function(arr, k, out):
            zlim, ylim, xlim = np.shape(arr)
            zp, yp, xp = np.shape(k)
            zp, yp, xp = int(zp / 2), int(yp / 2), int(xp / 2)
            sum = np.sum(k)
            sizep = k.size
            for z in prange(zlim) :
                for y in range(ylim) :
                    for x in range(xlim) :
                        zmin, zmax = max(0, z - zp), min(zlim, z + zp + 1)
                        ymin, ymax = max(0, y - yp), min(ylim, y + yp + 1)
                        xmin, xmax = max(0, x - xp), min(xlim, x + xp + 1)
                        n = (zmax - zmin) * (ymax - ymin) * (xmax - xmin)
                        if n != sizep :
                            ksum = np.sum(k[zmin - z + zp: zmax - z + zp, ymin - y + yp: ymax - y + yp, xmin - x + xp: xmax - x + xp])
                            if ksum != 0 :
                                n = n * sum / ksum
                        crop = arr[zmin: zmax, ymin: ymax, xmin: xmax] * k[zmin - z + zp: zmax - z + zp, ymin - y + yp: ymax - y + yp, xmin - x + xp: xmax - x + xp] * n
                        out[z, y, x] = func(crop)
        
    else :
        raise SyntaxError(f'Correlation error cannot be {method}')
    _cached[method] = function
    return function



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-31
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : relabel

"""
This file allows to test relabel

relabel : This function redefines a label array to fill potential holes inside.
"""



# %% Libraries
from corelp import print, debug
import pytest
from arrlp import relabel
debug_folder = debug(__file__)



# %% Function test
def test_function() :
    '''
    Test relabel function
    '''
    print('Hello world!')



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-31
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : relabel

"""
This function redefines a label array to fill potential holes inside.
"""



# %% Libraries

import numpy as np
from numba import njit

@njit()
def relabel(array):
    '''
    This function redefines a label array to fill potential holes inside.
    
    Parameters
    ----------
    labels : np.ndarray
        array of labels (ints where each number > 0 corresponds to one object).

    Returns
    -------
    labels : np.ndarray
        1D version of the corrected labels array.

    Examples
    --------
    >>> from arrlp import relabel
    ...
    >>> relabel(labels)
    '''

    array = array.ravel()  # Flatten array
    unique_values = np.unique(array)  # Get sorted unique values
    mapping = np.zeros(unique_values[-1] + 1, dtype=np.uint32)  # Create mapping array

    # Assign new sequential labels
    for new_label, old_value in enumerate(unique_values):
        mapping[old_value] = new_label  # Map old value to new one

    # Apply relabeling
    for i in range(len(array)):
        array[i] = mapping[array[i]]

    return array.reshape((-1,))  # Return as 1D array


# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : convolve

"""
This file allows to test convolve

convolve : This function convolves two numpy arrays (1D, 2D, 3D).
"""



# %% Libraries
from corelp import debug
import pytest
from matplotlib import pyplot as plt
from arrlp import kernel, convolve
debug_folder = debug(__file__)



# %% Function test
@pytest.mark.parametrize("kwargs, title", [
    ({"window":10}, "Mask"),
    ({"sigma":5}, "Gaussian"),
    ({"wl":640, "NA":1.5}, "Airy"),
])
def test_function(kwargs, title) :
    '''
    Test correlate values
    '''

    img = kernel(pixel=5, window=1000)
    
    plt.figure()
    plt.imshow(img, cmap="hot")
    plt.tight_layout()
    plt.colorbar()
    plt.savefig(debug_folder / f"{title}_before.png")

    out = convolve(img, pixel=5, **kwargs)

    plt.figure()
    plt.imshow(out, cmap="hot")
    plt.tight_layout()
    plt.colorbar()
    plt.savefig(debug_folder / f"{title}.png")




# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-30
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : coordinates

"""
This function will return a ndarray corresponding to the coordinates array of the input.
"""



# %% Libraries
import numpy as np



# %% Function
def coordinates(shape, pixel=1., *, ndims=1, center=True, grid=False, origin=0.) :
    '''
    This function will return a ndarray corresponding to the coordinates array of the input.
    
    Parameters
    ----------
    shape : int or tuple or np.ndarray
        Describes the shape of the coordinates.
        If int, all dimensions will have this value.
        If tuple, corresponds to the shape.
        If array, will take the shape attribute of the array
    pixel : float, tuple(float)
        Says what is the size of one bin, default is 1.
        If tuple, corresponds to the pixel for each dimension.
    ndims : int
        Number of dimensions if input shape is int.
        Will be overriden by any other input dimension.
    center : bool or tuple(bool)
        True to center origin, False to start by origin.
        If tuple, defines how to place origin for each dimension.
    grid : bool
        True to return full meshgrid for each dimensions, False for broadcastable arrays.
    origin : float or tuple(float)
        Value of origin (default 0).
        If tuple, value for each dimension.

    Returns
    -------
    coords : np.ndarray or tuple(np.ndarray)
        for 1 dimensional shape, returns coordinates array.
        else, return a tuple of arrays for each dimensions.

    Raises
    ------
    ValueError
        if input or passed as tuples that do not have the good number of dimensions.

    Examples
    --------
    >>> from arrlp import coordinates
    ...
    >>> array = np.ones((5,4))
    >>> coordinates = coordinates(array, pixel=10, center=(True, False), grid=False, origin=(0,3))
    ... array([[-20.], [-10.], [0.], [10.], [20.]]), array([[3., 13., 23., 33.]])
    '''

    # Manage shape argument
    if isinstance(shape, int) :
        shape = (shape,) * ndims
    elif isinstance(shape, tuple) :
        pass
    else :
        shape = np.shape(shape)

    # Correct ndims
    ndims = len(shape)

    # Correct input dimensions
    pixel = iterable(pixel, ndims)
    center = iterable(center, ndims)
    origin = iterable(origin, ndims)

    #looping on shape
    coords = []
    for dim, (n, pix, cent, orig) in enumerate(zip(shape, pixel, center, origin)) :
        coord = n2coord(n, pixel=pix, center=cent, origin=orig)
        reshape = np.ones(ndims, dtype=int)
        reshape[dim] = len(coord)
        coord = coord.reshape(tuple(reshape))
        coords.append(coord)

    #Meshgrid
    if ndims == 1 :
        return coords[0]
    elif grid :
        return np.meshgrid(*coords, indexing='ij')
    else :
        return coords



def iterable(arg, ndims) :
    try :
        if len(arg) != ndims :
            raise ValueError('Argument does not have good number of dimensions')
        else :
            return arg
    except TypeError :
        return (arg,) * ndims



def n2coord(n, pixel=1., center=False, origin=0.) :
    if center :
        start, stop = -(n - 1) / 2,(n - 1) / 2
    else :
        start,stop = 0, n - 1
    return np.linspace(start, stop, n) * pixel + origin



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)